"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import logging
from typing import Dict

from ..exceptions import (
    CertRecordParseException,
    RemoteCertificateExpired,
    TofuCertificateRejection,
)
from .cert_record import CertRecord
from .cert_wrapper import CertWrapper

logger = logging.getLogger(__name__)


class CertStore:
    """
    Data structure to store the certificates across visited hosts
    """

    __hosts_file: str
    __cert_store_data: Dict[str, CertRecord]

    def __init__(self, hosts_file):
        """
        Initializes a new cert store with a specified file to store the certificate fingerprint & expiration dates
        """
        self.__cert_store_data = {}
        self.__hosts_file = hosts_file

    def set_hosts_file(self, hosts_file):
        """
        Updates the specified file for certificate fingerprint storage
        """
        self.__hosts_file = hosts_file

    def get_hosts_file(self):
        """
        Returns the currently set hosts file location
        """
        return self.__hosts_file

    def validate_tofu_or_add(self, hostname: str, cert: CertWrapper) -> bool:
        """
        Given the hostname & correspoding certificate, this function:
        1. Checks to see if the certificate is expired (if so, it throws a RemoteCertificateExpired exception)
        2. Fetches a corresponding stored certificate record from the client to implement a TOFU check:
          a. If there is a local cert record, and it's not expired, and it matches the passed certificate, return success
          b. If there is not a local cert record, save the certificate record locally, and return success
          c. If there is a local cert record, but it's expired, save the certificate record locally, and return success
          d. If there is a local cert record, and it's not expired, but it does not match the passed certificate,
             throw TofuCertificateRejection
        """
        remote_cert_record = CertRecord(hostname, cert.fingerprint(), cert.expiration())

        if remote_cert_record.is_expired():
            raise RemoteCertificateExpired

        local_cert_record = self.__get_cert_record(hostname)

        if (
            local_cert_record
            and not local_cert_record.is_expired()
            and local_cert_record.fingerprint != remote_cert_record.fingerprint
        ):
            raise TofuCertificateRejection

        self.__add_cert_record(remote_cert_record)
        return True

    def __get_cert_record(self, hostname: str) -> CertRecord:
        """
        Fetch the corresponding CertRecord for passed hostname from the local storage (file)
        TODO: smarter loading logic
        """
        self.__load()
        return self.__cert_store_data.get(hostname, None)

    def __add_cert_record(self, cert_record: CertRecord):
        """
        Add a CertRecord for the corresponding hostname to local storage (file) and save to file
        TODO: smarter saving logic
        """
        self.__cert_store_data[cert_record.hostname] = cert_record
        self.__save()
        return self

    def __load(self):
        """
        Reloads the hosts file from storage and copies that into memory
        """
        file_lines = []
        try:
            with open(self.__hosts_file, "r", encoding="utf-8") as f:
                file_lines = f.readlines()
        except FileNotFoundError:
            file_lines = []

        for file_line in file_lines:
            cert_record = self.__load_record(file_line)
            if cert_record is not None:
                self.__cert_store_data[cert_record.hostname] = cert_record

        return self

    def __load_record(self, file_line):
        try:
            return CertRecord.from_string(file_line)
        except CertRecordParseException:
            logger.warning(
                f"Invalid TOFU record encountered: '{file_line.strip()}'. This record has been skipped."
            )
            return None

    def __save(self):
        """
        Saves the full set of host records back to file
        """
        with open(self.__hosts_file, "w", encoding="utf-8") as f:
            for c in self.__cert_store_data.values():
                f.write(c.to_string())

        return self



"""Parse (absolute and relative) URLs.

urlparse module is based upon the following RFC specifications.

RFC 3986 (STD66): "Uniform Resource Identifiers" by T. Berners-Lee, R. Fielding
and L.  Masinter, January 2005.

RFC 2732 : "Format for Literal IPv6 Addresses in URL's by R.Hinden, B.Carpenter
and L.Masinter, December 1999.

RFC 2396:  "Uniform Resource Identifiers (URI)": Generic Syntax by T.
Berners-Lee, R. Fielding, and L. Masinter, August 1998.

RFC 2368: "The mailto URL scheme", by P.Hoffman , L Masinter, J. Zawinski, July 1998.

RFC 1808: "Relative Uniform Resource Locators", by R. Fielding, UC Irvine, June
1995.

RFC 1738: "Uniform Resource Locators (URL)" by T. Berners-Lee, L. Masinter, M.
McCahill, December 1994

RFC 3986 is considered the current standard and any future changes to
urlparse module should conform with it.  The urlparse module is
currently not entirely compliant with this RFC due to defacto
scenarios for parsing, and for backward compatibility purposes, some
parsing quirks from older RFCs are retained. The testcases in
test_urlparse.py provides a good indicator of parsing behavior.
"""

import collections
import re
import sys
import warnings

__all__ = [
    "urlparse",
    "urlunparse",
    "urljoin",
    "urldefrag",
    "urlsplit",
    "urlunsplit",
    "urlencode",
    "parse_qs",
    "parse_qsl",
    "quote",
    "quote_plus",
    "quote_from_bytes",
    "unquote",
    "unquote_plus",
    "unquote_to_bytes",
    "DefragResult",
    "ParseResult",
    "SplitResult",
    "DefragResultBytes",
    "ParseResultBytes",
    "SplitResultBytes",
]

# A classification of schemes.
# The empty string classifies URLs with no scheme specified,
# being the default value returned by “urlsplit” and “urlparse”.

uses_relative = [
    "",
    "ftp",
    "http",
    "gemini",  # PATCH added gemini
    "gopher",
    "nntp",
    "imap",
    "wais",
    "file",
    "https",
    "shttp",
    "mms",
    "prospero",
    "rtsp",
    "rtspu",
    "sftp",
    "svn",
    "svn+ssh",
    "ws",
    "wss",
]

uses_netloc = [
    "",
    "ftp",
    "http",
    "gemini",  # PATCH added gemini
    "gopher",
    "nntp",
    "telnet",
    "imap",
    "wais",
    "file",
    "mms",
    "https",
    "shttp",
    "snews",
    "prospero",
    "rtsp",
    "rtspu",
    "rsync",
    "svn",
    "svn+ssh",
    "sftp",
    "nfs",
    "git",
    "git+ssh",
    "ws",
    "wss",
]

uses_params = [
    "",
    "ftp",
    "gemini",  # PATCH added gemini
    "hdl",
    "prospero",
    "http",
    "imap",
    "https",
    "shttp",
    "rtsp",
    "rtspu",
    "sip",
    "sips",
    "mms",
    "sftp",
    "tel",
]

# These are not actually used anymore, but should stay for backwards
# compatibility.  (They are undocumented, but have a public-looking name.)

non_hierarchical = [
    "gopher",
    "hdl",
    "mailto",
    "news",
    "telnet",
    "wais",
    "imap",
    "snews",
    "sip",
    "sips",
]

uses_query = [
    "",
    "http",
    "wais",
    "imap",
    "https",
    "shttp",
    "mms",
    "gopher",
    "rtsp",
    "rtspu",
    "sip",
    "sips",
]

uses_fragment = [
    "",
    "ftp",
    "hdl",
    "http",
    "gopher",
    "news",
    "nntp",
    "wais",
    "https",
    "shttp",
    "snews",
    "file",
    "prospero",
]

# Characters valid in scheme names
scheme_chars = (
    "abcdefghijklmnopqrstuvwxyz" "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789" "+-."
)

# XXX: Consider replacing with functools.lru_cache
MAX_CACHE_SIZE = 20
_parse_cache = {}


def clear_cache():
    """Clear the parse cache and the quoters cache."""
    _parse_cache.clear()
    _safe_quoters.clear()


# Helpers for bytes handling
# For 3.2, we deliberately require applications that
# handle improperly quoted URLs to do their own
# decoding and encoding. If valid use cases are
# presented, we may relax this by using latin-1
# decoding internally for 3.3
_implicit_encoding = "ascii"
_implicit_errors = "strict"


def _noop(obj):
    return obj


def _encode_result(obj, encoding=_implicit_encoding, errors=_implicit_errors):
    return obj.encode(encoding, errors)


def _decode_args(args, encoding=_implicit_encoding, errors=_implicit_errors):
    return tuple(x.decode(encoding, errors) if x else "" for x in args)


def _coerce_args(*args):
    # Invokes decode if necessary to create str args
    # and returns the coerced inputs along with
    # an appropriate result coercion function
    #   - noop for str inputs
    #   - encoding function otherwise
    str_input = isinstance(args[0], str)
    for arg in args[1:]:
        # We special-case the empty string to support the
        # "scheme=''" default argument to some functions
        if arg and isinstance(arg, str) != str_input:
            raise TypeError("Cannot mix str and non-str arguments")
    if str_input:
        return args + (_noop,)
    return _decode_args(args) + (_encode_result,)


# Result objects are more helpful than simple tuples
class _ResultMixinStr(object):
    """Standard approach to encoding parsed results from str to bytes"""

    __slots__ = ()

    def encode(self, encoding="ascii", errors="strict"):
        return self._encoded_counterpart(*(x.encode(encoding, errors) for x in self))


class _ResultMixinBytes(object):
    """Standard approach to decoding parsed results from bytes to str"""

    __slots__ = ()

    def decode(self, encoding="ascii", errors="strict"):
        return self._decoded_counterpart(*(x.decode(encoding, errors) for x in self))


class _NetlocResultMixinBase(object):
    """Shared methods for the parsed result objects containing a netloc element"""

    __slots__ = ()

    @property
    def username(self):
        return self._userinfo[0]

    @property
    def password(self):
        return self._userinfo[1]

    @property
    def hostname(self):
        hostname = self._hostinfo[0]
        if not hostname:
            return None
        # Scoped IPv6 address may have zone info, which must not be lowercased
        # like http://[fe80::822a:a8ff:fe49:470c%tESt]:1234/keys
        separator = "%" if isinstance(hostname, str) else b"%"
        hostname, percent, zone = hostname.partition(separator)
        return hostname.lower() + percent + zone

    @property
    def port(self):
        port = self._hostinfo[1]
        if port is not None:
            try:
                port = int(port, 10)
            except ValueError:
                message = f"Port could not be cast to integer value as {port!r}"
                raise ValueError(message) from None
            if not (0 <= port <= 65535):
                raise ValueError("Port out of range 0-65535")
        return port


class _NetlocResultMixinStr(_NetlocResultMixinBase, _ResultMixinStr):
    __slots__ = ()

    @property
    def _userinfo(self):
        netloc = self.netloc
        userinfo, have_info, hostinfo = netloc.rpartition("@")
        if have_info:
            username, have_password, password = userinfo.partition(":")
            if not have_password:
                password = None
        else:
            username = password = None
        return username, password

    @property
    def _hostinfo(self):
        netloc = self.netloc
        _, _, hostinfo = netloc.rpartition("@")
        _, have_open_br, bracketed = hostinfo.partition("[")
        if have_open_br:
            hostname, _, port = bracketed.partition("]")
            _, _, port = port.partition(":")
        else:
            hostname, _, port = hostinfo.partition(":")
        if not port:
            port = None
        return hostname, port


class _NetlocResultMixinBytes(_NetlocResultMixinBase, _ResultMixinBytes):
    __slots__ = ()

    @property
    def _userinfo(self):
        netloc = self.netloc
        userinfo, have_info, hostinfo = netloc.rpartition(b"@")
        if have_info:
            username, have_password, password = userinfo.partition(b":")
            if not have_password:
                password = None
        else:
            username = password = None
        return username, password

    @property
    def _hostinfo(self):
        netloc = self.netloc
        _, _, hostinfo = netloc.rpartition(b"@")
        _, have_open_br, bracketed = hostinfo.partition(b"[")
        if have_open_br:
            hostname, _, port = bracketed.partition(b"]")
            _, _, port = port.partition(b":")
        else:
            hostname, _, port = hostinfo.partition(b":")
        if not port:
            port = None
        return hostname, port


from collections import namedtuple

_DefragResultBase = namedtuple("DefragResult", "url fragment")
_SplitResultBase = namedtuple("SplitResult", "scheme netloc path query fragment")
_ParseResultBase = namedtuple("ParseResult", "scheme netloc path params query fragment")

_DefragResultBase.__doc__ = """
DefragResult(url, fragment)

A 2-tuple that contains the url without fragment identifier and the fragment
identifier as a separate argument.
"""

_DefragResultBase.url.__doc__ = """The URL with no fragment identifier."""

_DefragResultBase.fragment.__doc__ = """
Fragment identifier separated from URL, that allows indirect identification of a
secondary resource by reference to a primary resource and additional identifying
information.
"""

_SplitResultBase.__doc__ = """
SplitResult(scheme, netloc, path, query, fragment)

A 5-tuple that contains the different components of a URL. Similar to
ParseResult, but does not split params.
"""

_SplitResultBase.scheme.__doc__ = """Specifies URL scheme for the request."""

_SplitResultBase.netloc.__doc__ = """
Network location where the request is made to.
"""

_SplitResultBase.path.__doc__ = """
The hierarchical path, such as the path to a file to download.
"""

_SplitResultBase.query.__doc__ = """
The query component, that contains non-hierarchical data, that along with data
in path component, identifies a resource in the scope of URI's scheme and
network location.
"""

_SplitResultBase.fragment.__doc__ = """
Fragment identifier, that allows indirect identification of a secondary resource
by reference to a primary resource and additional identifying information.
"""

_ParseResultBase.__doc__ = """
ParseResult(scheme, netloc, path, params, query, fragment)

A 6-tuple that contains components of a parsed URL.
"""

_ParseResultBase.scheme.__doc__ = _SplitResultBase.scheme.__doc__
_ParseResultBase.netloc.__doc__ = _SplitResultBase.netloc.__doc__
_ParseResultBase.path.__doc__ = _SplitResultBase.path.__doc__
_ParseResultBase.params.__doc__ = """
Parameters for last path element used to dereference the URI in order to provide
access to perform some operation on the resource.
"""

_ParseResultBase.query.__doc__ = _SplitResultBase.query.__doc__
_ParseResultBase.fragment.__doc__ = _SplitResultBase.fragment.__doc__


# For backwards compatibility, alias _NetlocResultMixinStr
# ResultBase is no longer part of the documented API, but it is
# retained since deprecating it isn't worth the hassle
ResultBase = _NetlocResultMixinStr


# Structured result objects for string data
class DefragResult(_DefragResultBase, _ResultMixinStr):
    __slots__ = ()

    def geturl(self):
        if self.fragment:
            return self.url + "#" + self.fragment
        else:
            return self.url


class SplitResult(_SplitResultBase, _NetlocResultMixinStr):
    __slots__ = ()

    def geturl(self):
        return urlunsplit(self)


class ParseResult(_ParseResultBase, _NetlocResultMixinStr):
    __slots__ = ()

    def geturl(self):
        return urlunparse(self)


# Structured result objects for bytes data
class DefragResultBytes(_DefragResultBase, _ResultMixinBytes):
    __slots__ = ()

    def geturl(self):
        if self.fragment:
            return self.url + b"#" + self.fragment
        else:
            return self.url


class SplitResultBytes(_SplitResultBase, _NetlocResultMixinBytes):
    __slots__ = ()

    def geturl(self):
        return urlunsplit(self)


class ParseResultBytes(_ParseResultBase, _NetlocResultMixinBytes):
    __slots__ = ()

    def geturl(self):
        return urlunparse(self)


# Set up the encode/decode result pairs
def _fix_result_transcoding():
    _result_pairs = (
        (DefragResult, DefragResultBytes),
        (SplitResult, SplitResultBytes),
        (ParseResult, ParseResultBytes),
    )
    for _decoded, _encoded in _result_pairs:
        _decoded._encoded_counterpart = _encoded
        _encoded._decoded_counterpart = _decoded


_fix_result_transcoding()
del _fix_result_transcoding


def urlparse(url, scheme="", allow_fragments=True):
    """Parse a URL into 6 components:
    <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
    Return a 6-tuple: (scheme, netloc, path, params, query, fragment).
    Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes."""
    url, scheme, _coerce_result = _coerce_args(url, scheme)
    splitresult = urlsplit(url, scheme, allow_fragments)
    scheme, netloc, url, query, fragment = splitresult
    if scheme in uses_params and ";" in url:
        url, params = _splitparams(url)
    else:
        params = ""
    result = ParseResult(scheme, netloc, url, params, query, fragment)
    return _coerce_result(result)


def _splitparams(url):
    if "/" in url:
        i = url.find(";", url.rfind("/"))
        if i < 0:
            return url, ""
    else:
        i = url.find(";")
    return url[:i], url[i + 1 :]


def _splitnetloc(url, start=0):
    delim = len(url)  # position of end of domain part of url, default is end
    for c in "/?#":  # look for delimiters; the order is NOT important
        wdelim = url.find(c, start)  # find first of this delim
        if wdelim >= 0:  # if found
            delim = min(delim, wdelim)  # use earliest delim position
    return url[start:delim], url[delim:]  # return (domain, rest)


def _checknetloc(netloc):
    if not netloc or netloc.isascii():
        return
    # looking for characters like \u2100 that expand to 'a/c'
    # IDNA uses NFKC equivalence, so normalize for this check
    import unicodedata

    n = netloc.replace("@", "")  # ignore characters already included
    n = n.replace(":", "")  # but not the surrounding text
    n = n.replace("#", "")
    n = n.replace("?", "")
    netloc2 = unicodedata.normalize("NFKC", n)
    if n == netloc2:
        return
    for c in "/?#@:":
        if c in netloc2:
            raise ValueError(
                "netloc '"
                + netloc
                + "' contains invalid "
                + "characters under NFKC normalization"
            )


def urlsplit(url, scheme="", allow_fragments=True):
    """Parse a URL into 5 components:
    <scheme>://<netloc>/<path>?<query>#<fragment>
    Return a 5-tuple: (scheme, netloc, path, query, fragment).
    Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes."""
    url, scheme, _coerce_result = _coerce_args(url, scheme)
    allow_fragments = bool(allow_fragments)
    key = url, scheme, allow_fragments, type(url), type(scheme)
    cached = _parse_cache.get(key, None)
    if cached:
        return _coerce_result(cached)
    if len(_parse_cache) >= MAX_CACHE_SIZE:  # avoid runaway growth
        clear_cache()
    netloc = query = fragment = ""
    i = url.find(":")
    if i > 0:
        if url[:i] == "http":  # optimize the common case
            url = url[i + 1 :]
            if url[:2] == "//":
                netloc, url = _splitnetloc(url, 2)
                if ("[" in netloc and "]" not in netloc) or (
                    "]" in netloc and "[" not in netloc
                ):
                    raise ValueError("Invalid IPv6 URL")
            if allow_fragments and "#" in url:
                url, fragment = url.split("#", 1)
            if "?" in url:
                url, query = url.split("?", 1)
            _checknetloc(netloc)
            v = SplitResult("http", netloc, url, query, fragment)
            _parse_cache[key] = v
            return _coerce_result(v)
        for c in url[:i]:
            if c not in scheme_chars:
                break
        else:
            # make sure "url" is not actually a port number (in which case
            # "scheme" is really part of the path)
            rest = url[i + 1 :]
            if not rest or any(c not in "0123456789" for c in rest):
                # not a port number
                scheme, url = url[:i].lower(), rest

    if url[:2] == "//":
        netloc, url = _splitnetloc(url, 2)
        if ("[" in netloc and "]" not in netloc) or (
            "]" in netloc and "[" not in netloc
        ):
            raise ValueError("Invalid IPv6 URL")
    if allow_fragments and "#" in url:
        url, fragment = url.split("#", 1)
    if "?" in url:
        url, query = url.split("?", 1)
    _checknetloc(netloc)
    v = SplitResult(scheme, netloc, url, query, fragment)
    _parse_cache[key] = v
    return _coerce_result(v)


def urlunparse(components):
    """Put a parsed URL back together again.  This may result in a
    slightly different, but equivalent URL, if the URL that was parsed
    originally had redundant delimiters, e.g. a ? with an empty query
    (the draft states that these are equivalent)."""
    scheme, netloc, url, params, query, fragment, _coerce_result = _coerce_args(
        *components
    )
    if params:
        url = "%s;%s" % (url, params)
    return _coerce_result(urlunsplit((scheme, netloc, url, query, fragment)))


def urlunsplit(components):
    """Combine the elements of a tuple as returned by urlsplit() into a
    complete URL as a string. The data argument can be any five-item iterable.
    This may result in a slightly different, but equivalent URL, if the URL that
    was parsed originally had unnecessary delimiters (for example, a ? with an
    empty query; the RFC states that these are equivalent)."""
    scheme, netloc, url, query, fragment, _coerce_result = _coerce_args(*components)
    if netloc or (scheme and scheme in uses_netloc and url[:2] != "//"):
        if url and url[:1] != "/":
            url = "/" + url
        url = "//" + (netloc or "") + url
    if scheme:
        url = scheme + ":" + url
    if query:
        url = url + "?" + query
    if fragment:
        url = url + "#" + fragment
    return _coerce_result(url)


def urljoin(base, url, allow_fragments=True):
    """Join a base URL and a possibly relative URL to form an absolute
    interpretation of the latter."""
    if not base:
        return url
    if not url:
        return base

    base, url, _coerce_result = _coerce_args(base, url)
    bscheme, bnetloc, bpath, bparams, bquery, bfragment = urlparse(
        base, "", allow_fragments
    )
    scheme, netloc, path, params, query, fragment = urlparse(
        url, bscheme, allow_fragments
    )

    if scheme != bscheme or scheme not in uses_relative:
        return _coerce_result(url)
    if scheme in uses_netloc:
        if netloc:
            return _coerce_result(
                urlunparse((scheme, netloc, path, params, query, fragment))
            )
        netloc = bnetloc

    if not path and not params:
        path = bpath
        params = bparams
        if not query:
            query = bquery
        return _coerce_result(
            urlunparse((scheme, netloc, path, params, query, fragment))
        )

    base_parts = bpath.split("/")
    if base_parts[-1] != "":
        # the last item is not a directory, so will not be taken into account
        # in resolving the relative path
        del base_parts[-1]

    # for rfc3986, ignore all base path should the first character be root.
    if path[:1] == "/":
        segments = path.split("/")
    else:
        segments = base_parts + path.split("/")
        # filter out elements that would cause redundant slashes on re-joining
        # the resolved_path
        segments[1:-1] = filter(None, segments[1:-1])

    resolved_path = []

    for seg in segments:
        if seg == "..":
            try:
                resolved_path.pop()
            except IndexError:
                # ignore any .. segments that would otherwise cause an IndexError
                # when popped from resolved_path if resolving for rfc3986
                pass
        elif seg == ".":
            continue
        else:
            resolved_path.append(seg)

    if segments[-1] in (".", ".."):
        # do some post-processing here. if the last segment was a relative dir,
        # then we need to append the trailing '/'
        resolved_path.append("")

    return _coerce_result(
        urlunparse(
            (scheme, netloc, "/".join(resolved_path) or "/", params, query, fragment)
        )
    )


def urldefrag(url):
    """Removes any existing fragment from URL.

    Returns a tuple of the defragmented URL and the fragment.  If
    the URL contained no fragments, the second element is the
    empty string.
    """
    url, _coerce_result = _coerce_args(url)
    if "#" in url:
        s, n, p, a, q, frag = urlparse(url)
        defrag = urlunparse((s, n, p, a, q, ""))
    else:
        frag = ""
        defrag = url
    return _coerce_result(DefragResult(defrag, frag))


_hexdig = "0123456789ABCDEFabcdef"
_hextobyte = None


def unquote_to_bytes(string):
    """unquote_to_bytes('abc%20def') -> b'abc def'."""
    # Note: strings are encoded as UTF-8. This is only an issue if it contains
    # unescaped non-ASCII characters, which URIs should not.
    if not string:
        # Is it a string-like object?
        string.split
        return b""
    if isinstance(string, str):
        string = string.encode("utf-8")
    bits = string.split(b"%")
    if len(bits) == 1:
        return string
    res = [bits[0]]
    append = res.append
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    global _hextobyte
    if _hextobyte is None:
        _hextobyte = {
            (a + b).encode(): bytes.fromhex(a + b) for a in _hexdig for b in _hexdig
        }
    for item in bits[1:]:
        try:
            append(_hextobyte[item[:2]])
            append(item[2:])
        except KeyError:
            append(b"%")
            append(item)
    return b"".join(res)


_asciire = re.compile("([\x00-\x7f]+)")


def unquote(string, encoding="utf-8", errors="replace"):
    """Replace %xx escapes by their single-character equivalent. The optional
    encoding and errors parameters specify how to decode percent-encoded
    sequences into Unicode characters, as accepted by the bytes.decode()
    method.
    By default, percent-encoded sequences are decoded with UTF-8, and invalid
    sequences are replaced by a placeholder character.

    unquote('abc%20def') -> 'abc def'.
    """
    if isinstance(string, bytes):
        raise TypeError("Expected str, got bytes")
    if "%" not in string:
        string.split
        return string
    if encoding is None:
        encoding = "utf-8"
    if errors is None:
        errors = "replace"
    bits = _asciire.split(string)
    res = [bits[0]]
    append = res.append
    for i in range(1, len(bits), 2):
        append(unquote_to_bytes(bits[i]).decode(encoding, errors))
        append(bits[i + 1])
    return "".join(res)


def parse_qs(
    qs,
    keep_blank_values=False,
    strict_parsing=False,
    encoding="utf-8",
    errors="replace",
    max_num_fields=None,
):
    """Parse a query given as a string argument.

    Arguments:

    qs: percent-encoded query string to be parsed

    keep_blank_values: flag indicating whether blank values in
        percent-encoded queries should be treated as blank strings.
        A true value indicates that blanks should be retained as
        blank strings.  The default false value indicates that
        blank values are to be ignored and treated as if they were
        not included.

    strict_parsing: flag indicating what to do with parsing errors.
        If false (the default), errors are silently ignored.
        If true, errors raise a ValueError exception.

    encoding and errors: specify how to decode percent-encoded sequences
        into Unicode characters, as accepted by the bytes.decode() method.

    max_num_fields: int. If set, then throws a ValueError if there
        are more than n fields read by parse_qsl().

    Returns a dictionary.
    """
    parsed_result = {}
    pairs = parse_qsl(
        qs,
        keep_blank_values,
        strict_parsing,
        encoding=encoding,
        errors=errors,
        max_num_fields=max_num_fields,
    )
    for name, value in pairs:
        if name in parsed_result:
            parsed_result[name].append(value)
        else:
            parsed_result[name] = [value]
    return parsed_result


def parse_qsl(
    qs,
    keep_blank_values=False,
    strict_parsing=False,
    encoding="utf-8",
    errors="replace",
    max_num_fields=None,
):
    """Parse a query given as a string argument.

    Arguments:

    qs: percent-encoded query string to be parsed

    keep_blank_values: flag indicating whether blank values in
        percent-encoded queries should be treated as blank strings.
        A true value indicates that blanks should be retained as blank
        strings.  The default false value indicates that blank values
        are to be ignored and treated as if they were  not included.

    strict_parsing: flag indicating what to do with parsing errors. If
        false (the default), errors are silently ignored. If true,
        errors raise a ValueError exception.

    encoding and errors: specify how to decode percent-encoded sequences
        into Unicode characters, as accepted by the bytes.decode() method.

    max_num_fields: int. If set, then throws a ValueError
        if there are more than n fields read by parse_qsl().

    Returns a list, as G-d intended.
    """
    qs, _coerce_result = _coerce_args(qs)

    # If max_num_fields is defined then check that the number of fields
    # is less than max_num_fields. This prevents a memory exhaustion DOS
    # attack via post bodies with many fields.
    if max_num_fields is not None:
        num_fields = 1 + qs.count("&") + qs.count(";")
        if max_num_fields < num_fields:
            raise ValueError("Max number of fields exceeded")

    pairs = [s2 for s1 in qs.split("&") for s2 in s1.split(";")]
    r = []
    for name_value in pairs:
        if not name_value and not strict_parsing:
            continue
        nv = name_value.split("=", 1)
        if len(nv) != 2:
            if strict_parsing:
                raise ValueError("bad query field: %r" % (name_value,))
            # Handle case of a control-name with no equal sign
            if keep_blank_values:
                nv.append("")
            else:
                continue
        if len(nv[1]) or keep_blank_values:
            name = nv[0].replace("+", " ")
            name = unquote(name, encoding=encoding, errors=errors)
            name = _coerce_result(name)
            value = nv[1].replace("+", " ")
            value = unquote(value, encoding=encoding, errors=errors)
            value = _coerce_result(value)
            r.append((name, value))
    return r


def unquote_plus(string, encoding="utf-8", errors="replace"):
    """Like unquote(), but also replace plus signs by spaces, as required for
    unquoting HTML form values.

    unquote_plus('%7e/abc+def') -> '~/abc def'
    """
    string = string.replace("+", " ")
    return unquote(string, encoding, errors)


_ALWAYS_SAFE = frozenset(
    b"ABCDEFGHIJKLMNOPQRSTUVWXYZ" b"abcdefghijklmnopqrstuvwxyz" b"0123456789" b"_.-~"
)
_ALWAYS_SAFE_BYTES = bytes(_ALWAYS_SAFE)
_safe_quoters = {}


class Quoter(collections.defaultdict):
    """A mapping from bytes (in range(0,256)) to strings.

    String values are percent-encoded byte values, unless the key < 128, and
    in the "safe" set (either the specified safe set, or default set).
    """

    # Keeps a cache internally, using defaultdict, for efficiency (lookups
    # of cached keys don't call Python code at all).
    def __init__(self, safe):
        """safe: bytes object."""
        self.safe = _ALWAYS_SAFE.union(safe)

    def __repr__(self):
        # Without this, will just display as a defaultdict
        return "<%s %r>" % (self.__class__.__name__, dict(self))

    def __missing__(self, b):
        # Handle a cache miss. Store quoted string in cache and return.
        res = chr(b) if b in self.safe else "%{:02X}".format(b)
        self[b] = res
        return res


def quote(string, safe="/", encoding=None, errors=None):
    """quote('abc def') -> 'abc%20def'

    Each part of a URL, e.g. the path info, the query, etc., has a
    different set of reserved characters that must be quoted. The
    quote function offers a cautious (not minimal) way to quote a
    string for most of these parts.

    RFC 3986 Uniform Resource Identifier (URI): Generic Syntax lists
    the following (un)reserved characters.

    unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
    reserved      = gen-delims / sub-delims
    gen-delims    = ":" / "/" / "?" / "#" / "[" / "]" / "@"
    sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
                  / "*" / "+" / "," / ";" / "="

    Each of the reserved characters is reserved in some component of a URL,
    but not necessarily in all of them.

    The quote function %-escapes all characters that are neither in the
    unreserved chars ("always safe") nor the additional chars set via the
    safe arg.

    The default for the safe arg is '/'. The character is reserved, but in
    typical usage the quote function is being called on a path where the
    existing slash characters are to be preserved.

    Python 3.7 updates from using RFC 2396 to RFC 3986 to quote URL strings.
    Now, "~" is included in the set of unreserved characters.

    string and safe may be either str or bytes objects. encoding and errors
    must not be specified if string is a bytes object.

    The optional encoding and errors parameters specify how to deal with
    non-ASCII characters, as accepted by the str.encode method.
    By default, encoding='utf-8' (characters are encoded with UTF-8), and
    errors='strict' (unsupported characters raise a UnicodeEncodeError).
    """
    if isinstance(string, str):
        if not string:
            return string
        if encoding is None:
            encoding = "utf-8"
        if errors is None:
            errors = "strict"
        string = string.encode(encoding, errors)
    else:
        if encoding is not None:
            raise TypeError("quote() doesn't support 'encoding' for bytes")
        if errors is not None:
            raise TypeError("quote() doesn't support 'errors' for bytes")
    return quote_from_bytes(string, safe)


def quote_plus(string, safe="", encoding=None, errors=None):
    """Like quote(), but also replace ' ' with '+', as required for quoting
    HTML form values. Plus signs in the original string are escaped unless
    they are included in safe. It also does not have safe default to '/'.
    """
    # Check if ' ' in string, where string may either be a str or bytes.  If
    # there are no spaces, the regular quote will produce the right answer.
    if (isinstance(string, str) and " " not in string) or (
        isinstance(string, bytes) and b" " not in string
    ):
        return quote(string, safe, encoding, errors)
    if isinstance(safe, str):
        space = " "
    else:
        space = b" "
    string = quote(string, safe + space, encoding, errors)
    return string.replace(" ", "+")


def quote_from_bytes(bs, safe="/"):
    """Like quote(), but accepts a bytes object rather than a str, and does
    not perform string-to-bytes encoding.  It always returns an ASCII string.
    quote_from_bytes(b'abc def\x3f') -> 'abc%20def%3f'
    """
    if not isinstance(bs, (bytes, bytearray)):
        raise TypeError("quote_from_bytes() expected bytes")
    if not bs:
        return ""
    if isinstance(safe, str):
        # Normalize 'safe' by converting to bytes and removing non-ASCII chars
        safe = safe.encode("ascii", "ignore")
    else:
        safe = bytes([c for c in safe if c < 128])
    if not bs.rstrip(_ALWAYS_SAFE_BYTES + safe):
        return bs.decode()
    try:
        quoter = _safe_quoters[safe]
    except KeyError:
        _safe_quoters[safe] = quoter = Quoter(safe).__getitem__
    return "".join([quoter(char) for char in bs])


def urlencode(
    query, doseq=False, safe="", encoding=None, errors=None, quote_via=quote_plus
):
    """Encode a dict or sequence of two-element tuples into a URL query string.

    If any values in the query arg are sequences and doseq is true, each
    sequence element is converted to a separate parameter.

    If the query arg is a sequence of two-element tuples, the order of the
    parameters in the output will match the order of parameters in the
    input.

    The components of a query arg may each be either a string or a bytes type.

    The safe, encoding, and errors parameters are passed down to the function
    specified by quote_via (encoding and errors only if a component is a str).
    """

    if hasattr(query, "items"):
        query = query.items()
    else:
        # It's a bother at times that strings and string-like objects are
        # sequences.
        try:
            # non-sequence items should not work with len()
            # non-empty strings will fail this
            if len(query) and not isinstance(query[0], tuple):
                raise TypeError
            # Zero-length sequences of all types will get here and succeed,
            # but that's a minor nit.  Since the original implementation
            # allowed empty dicts that type of behavior probably should be
            # preserved for consistency
        except TypeError:
            ty, va, tb = sys.exc_info()
            raise TypeError(
                "not a valid non-string sequence " "or mapping object"
            ).with_traceback(tb)

    l = []
    if not doseq:
        for k, v in query:
            if isinstance(k, bytes):
                k = quote_via(k, safe)
            else:
                k = quote_via(str(k), safe, encoding, errors)

            if isinstance(v, bytes):
                v = quote_via(v, safe)
            else:
                v = quote_via(str(v), safe, encoding, errors)
            l.append(k + "=" + v)
    else:
        for k, v in query:
            if isinstance(k, bytes):
                k = quote_via(k, safe)
            else:
                k = quote_via(str(k), safe, encoding, errors)

            if isinstance(v, bytes):
                v = quote_via(v, safe)
                l.append(k + "=" + v)
            elif isinstance(v, str):
                v = quote_via(v, safe, encoding, errors)
                l.append(k + "=" + v)
            else:
                try:
                    # Is this a sufficient test for sequence-ness?
                    x = len(v)
                except TypeError:
                    # not a sequence
                    v = quote_via(str(v), safe, encoding, errors)
                    l.append(k + "=" + v)
                else:
                    # loop over the sequence
                    for elt in v:
                        if isinstance(elt, bytes):
                            elt = quote_via(elt, safe)
                        else:
                            elt = quote_via(str(elt), safe, encoding, errors)
                        l.append(k + "=" + elt)
    return "&".join(l)


def to_bytes(url):
    warnings.warn(
        "urllib.parse.to_bytes() is deprecated as of 3.8",
        DeprecationWarning,
        stacklevel=2,
    )
    return _to_bytes(url)


def _to_bytes(url):
    """to_bytes(u"URL") --> 'URL'."""
    # Most URL schemes require ASCII. If that changes, the conversion
    # can be relaxed.
    # XXX get rid of to_bytes()
    if isinstance(url, str):
        try:
            url = url.encode("ASCII").decode()
        except UnicodeError:
            raise UnicodeError("URL " + repr(url) + " contains non-ASCII characters")
    return url


def unwrap(url):
    """Transform a string like '<URL:scheme://host/path>' into 'scheme://host/path'.

    The string is returned unchanged if it's not a wrapped URL.
    """
    url = str(url).strip()
    if url[:1] == "<" and url[-1:] == ">":
        url = url[1:-1].strip()
    if url[:4] == "URL:":
        url = url[4:].strip()
    return url


def splittype(url):
    warnings.warn(
        "urllib.parse.splittype() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splittype(url)


_typeprog = None


def _splittype(url):
    """splittype('type:opaquestring') --> 'type', 'opaquestring'."""
    global _typeprog
    if _typeprog is None:
        _typeprog = re.compile("([^/:]+):(.*)", re.DOTALL)

    match = _typeprog.match(url)
    if match:
        scheme, data = match.groups()
        return scheme.lower(), data
    return None, url


def splithost(url):
    warnings.warn(
        "urllib.parse.splithost() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splithost(url)


_hostprog = None


def _splithost(url):
    """splithost('//host[:port]/path') --> 'host[:port]', '/path'."""
    global _hostprog
    if _hostprog is None:
        _hostprog = re.compile("//([^/#?]*)(.*)", re.DOTALL)

    match = _hostprog.match(url)
    if match:
        host_port, path = match.groups()
        if path and path[0] != "/":
            path = "/" + path
        return host_port, path
    return None, url


def splituser(host):
    warnings.warn(
        "urllib.parse.splituser() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splituser(host)


def _splituser(host):
    """splituser('user[:passwd]@host[:port]') --> 'user[:passwd]', 'host[:port]'."""
    user, delim, host = host.rpartition("@")
    return (user if delim else None), host


def splitpasswd(user):
    warnings.warn(
        "urllib.parse.splitpasswd() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splitpasswd(user)


def _splitpasswd(user):
    """splitpasswd('user:passwd') -> 'user', 'passwd'."""
    user, delim, passwd = user.partition(":")
    return user, (passwd if delim else None)


def splitport(host):
    warnings.warn(
        "urllib.parse.splitport() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splitport(host)


# splittag('/path#tag') --> '/path', 'tag'
_portprog = None


def _splitport(host):
    """splitport('host:port') --> 'host', 'port'."""
    global _portprog
    if _portprog is None:
        _portprog = re.compile("(.*):([0-9]*)", re.DOTALL)

    match = _portprog.fullmatch(host)
    if match:
        host, port = match.groups()
        if port:
            return host, port
    return host, None


def splitnport(host, defport=-1):
    warnings.warn(
        "urllib.parse.splitnport() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splitnport(host, defport)


def _splitnport(host, defport=-1):
    """Split host and port, returning numeric port.
    Return given default port if no ':' found; defaults to -1.
    Return numerical port if a valid number are found after ':'.
    Return None if ':' but not a valid number."""
    host, delim, port = host.rpartition(":")
    if not delim:
        host = port
    elif port:
        try:
            nport = int(port)
        except ValueError:
            nport = None
        return host, nport
    return host, defport


def splitquery(url):
    warnings.warn(
        "urllib.parse.splitquery() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splitquery(url)


def _splitquery(url):
    """splitquery('/path?query') --> '/path', 'query'."""
    path, delim, query = url.rpartition("?")
    if delim:
        return path, query
    return url, None


def splittag(url):
    warnings.warn(
        "urllib.parse.splittag() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splittag(url)


def _splittag(url):
    """splittag('/path#tag') --> '/path', 'tag'."""
    path, delim, tag = url.rpartition("#")
    if delim:
        return path, tag
    return url, None


def splitattr(url):
    warnings.warn(
        "urllib.parse.splitattr() is deprecated as of 3.8, "
        "use urllib.parse.urlparse() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splitattr(url)


def _splitattr(url):
    """splitattr('/path;attr1=value1;attr2=value2;...') ->
    '/path', ['attr1=value1', 'attr2=value2', ...]."""
    words = url.split(";")
    return words[0], words[1:]


def splitvalue(attr):
    warnings.warn(
        "urllib.parse.splitvalue() is deprecated as of 3.8, "
        "use urllib.parse.parse_qsl() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return _splitvalue(attr)


def _splitvalue(attr):
    """splitvalue('attr=value') --> 'attr', 'value'."""
    attr, delim, value = attr.partition("=")
    return attr, (value if delim else None)



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""


class RemoteCertificateExpired(Exception):
    """
    An exception type to handle expired certificates from the remote server.
    This should throw if the remote certificate expiration date
    """


class TofuCertificateRejection(Exception):
    """
    An exception type handle TOFU (trust-on-first-use rejection).
    """


class CertRecordParseException(Exception):
    """
    An exception triggered on cert record parsing.
    """


class GeminiResponseParseError(Exception):
    """
    Raised when the gemini protocol data response cannot be parsed.
    """



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import cryptography
from cryptography import x509
from cryptography.hazmat.backends import default_backend


class CertWrapper:
    """
    Certificate as defined by x509
    """

    certificate: cryptography.x509.Certificate
    """
  Certificate fingerprint, to be used in TOFU handling and response
  """
    public_key_fingerprint: str

    def __init__(self, certificate: cryptography.x509.Certificate):
        """
        Constructor
        """
        self.certificate = certificate

    def expiration(self) -> str:
        """
        Access function for certificate expiration date
        """
        return self.certificate.not_valid_after

    def fingerprint(self) -> str:
        """
        Extracts the public key & expiration date from the cert,
        and returns the public key openssh fingerprint
        """
        return (
            self.certificate.public_key()
            .public_bytes(
                cryptography.hazmat.primitives.serialization.Encoding.OpenSSH,
                cryptography.hazmat.primitives.serialization.PublicFormat.OpenSSH,
            )
            .decode("utf-8")
        )

    @classmethod
    def parse(cls, raw_certificate: bytes):
        """
        Takes as input the raw certificate (originally from the TCP socket)
        Returns a certificate wrapper
        """
        x509_certificate = x509.load_der_x509_certificate(
            raw_certificate, default_backend()
        )
        return CertWrapper(x509_certificate)



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import logging

logger = logging.getLogger(__name__)


def normalize_path(path: str) -> str:
    """
    Implements a normalized path for a string
    See RFC-3986 https://tools.ietf.org/html/rfc3986#section-5.2.4
    5.2.4.  Remove Dot Segments

    Example:
      STEP   OUTPUT BUFFER         INPUT BUFFER

        1 :                         /a/b/c/./../../g
        2E:   /a                    /b/c/./../../g
        2E:   /a/b                  /c/./../../g
        2E:   /a/b/c                /./../../g
        2B:   /a/b/c                /../../g
        2C:   /a/b                  /../g
        2C:   /a                    /g
        2E:   /a/g
    """
    result_stack = []
    for component in path.split("/"):
        if component in (".", ""):
            continue  # Do nothing
        if component == "..":
            if len(result_stack) > 0:
                result_stack.pop()
        else:
            result_stack.append(component)

    unescaped_path = "".join(
        [
            ("/" if len(path) > 0 and path[0] == "/" else ""),
            ("/".join(result_stack)),
            ("/" if len(path) > 0 and path[len(path) - 1] == "/" else ""),
        ]
    )

    return unescaped_path.replace("//", "/")


class TimeoutManager:
    """
    Timeout Manager for global timeout management at the top-level
    """

    def __init__(self, default_timeout):
        """
        Sets a default timeout on initialization
        """
        self.set_default_timeout(default_timeout)

    def set_default_timeout(self, default_timeout):
        """
        Allow the default timeout to be overwritten
        """
        self.default_timeout = default_timeout

    def get_timeout(self, timeout):
        """
        Takes in a timeout and returns that, or the default timeout
        """
        if timeout is not None:
            return timeout
        return self.default_timeout



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import logging
import re
import socket
import ssl
from socket import gaierror as SocketGaiErrorException
from socket import herror as SocketHErrorException
from socket import timeout as SocketTimeoutException

import cryptography

from .exceptions import (
    GeminiResponseParseError,
    RemoteCertificateExpired,
    TofuCertificateRejection,
)
from .globals import (
    CRLF,
    GEMINI_DEFAULT_ENCODING,
    GEMINI_RESPONSE_HEADER_META_MAXLENGTH,
    GEMINI_RESPONSE_HEADER_SEPARATOR,
    RESPONSE_STATUSDETAIL_ERROR_DNS,
    RESPONSE_STATUSDETAIL_ERROR_HOST,
    RESPONSE_STATUSDETAIL_ERROR_PROTOCOL,
    RESPONSE_STATUSDETAIL_ERROR_TLS,
)
from .response import BaseResponse, ResponseFactory
from .ssl.cert_wrapper import CertWrapper
from .url import URL

logger = logging.getLogger(__name__)


class Request:
    """
    Handles a single request to a Gemini Server.

    The request handler has four key responsibilities:

    1. It manages resolution of the requested URL for
       the remote server, by invoking underlying URL parse
       logic
    2. It manages transmission of the request via TLS over a
       socket connection to the remote server.
    3. It validates SSL certificate response using a TOFU
       (trust-on-first-use) validation paradigm
    4. It manages raw response handling and designation to
       the Response object (or exception raising, if indicated)
    """

    def __init__(
        self,
        url: str,
        raise_errors=False,
        referer=None,
        request_timeout=None,
        cert_store=None,
        ca_cert=None,
    ):
        """
        Initializes Response with a url, referer, and timeout
        """

        self.__url = URL(url, referer_url=referer)
        self.__raise_errors = raise_errors
        self.__timeout = request_timeout
        self.__cert_store = cert_store
        self.__ca_cert = ca_cert  # This should be a tuple

    def get_url(self):
        """
        Fetch the generated URL for the request (based on referer, if present)
        """

        return str(self.__url)

    def send(self):
        """
        Performes network communication and returns a Response object
        """

        logger.debug(f"Attempting to create a connection to {self.__url.netloc()}")
        socket_result = self.__get_socket()
        if isinstance(socket_result, BaseResponse):
            return socket_result

        logger.debug(
            f"Attempting to negotiate SSL handshake with {self.__url.netloc()}"
        )
        secure_socket_result = self.__negotiate_ssl(socket_result)
        if isinstance(secure_socket_result, BaseResponse):
            return secure_socket_result

        logger.debug(f"Validating server certificate to {self.__url.netloc()}")
        ssl_certificate_result = self.__validate_ssl_certificate(secure_socket_result)
        if isinstance(ssl_certificate_result, BaseResponse):
            return ssl_certificate_result

        logger.debug(f"Sending request header: {self.__url}")
        transport_result = self.__transport_payload(secure_socket_result, self.__url)
        if isinstance(transport_result, BaseResponse):
            return transport_result

        header, raw_body = transport_result
        logger.debug(
            f"Received response header: [{header}] and payload of length {len(raw_body)} bytes"
        )
        return self.__handle_response(
            header, raw_body, ssl_certificate_result.certificate
        )

    def __get_socket(self):
        """
        Creates a socket connection and manages exceptions.
        """

        try:
            sock = socket.create_connection(
                (self.__url.host(), self.__url.port()), timeout=self.__timeout
            )
            logger.debug(f"Created socket connection: {sock}")
            return sock
        except ConnectionRefusedError as err:
            logger.debug(
                f"ConnectionRefusedError: Connection to {self.__url.netloc()} was refused. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_HOST, "Connection refused"
            )
        except ConnectionResetError as err:
            logger.debug(
                f"ConnectionResetError: Connection to {self.__url.netloc()} was reset. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_HOST, "Connection reset"
            )
        except SocketHErrorException as err:
            logger.debug(
                f"socket.herror: socket.gethostbyaddr returned for {self.__url.host()}. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_HOST, "Host error"
            )
        except SocketGaiErrorException as err:
            logger.debug(
                f"socket.gaierror: socket.getaddrinfo returned unknown host for {self.__url.host()}. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_DNS, "Unknown host"
            )
        except SocketTimeoutException as err:
            logger.debug(
                f"socket.timeout: socket timed out connecting to {self.__url.host()}. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_HOST, "Socket timeout"
            )
        except Exception as err:
            logger.error(
                f"Unknown exception encountered when connecting to {self.__url.netloc()} - {err}"
            )
            raise err

    def __negotiate_ssl(self, socket_obj) -> ssl.SSLSocket:
        """
        Negotiates a SSL handshake on the passed socket connection and returns the secure socket
        """

        try:
            context = self.__setup_ssl_default_context()

            if self.is_using_ca_cert():
                self.__setup_ssl_client_certificate_context(context)

            secure_socket_result = context.wrap_socket(
                socket_obj, server_hostname=self.__url.host()
            )
            return secure_socket_result
        except ssl.SSLZeroReturnError as err:
            logger.debug(f"ssl.SSLZeroReturnError for {self.__url.host()} - {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "SSL Zero Return Error"
            )
        except ssl.SSLWantReadError as err:
            logger.debug(f"ssl.SSLWantReadError for {self.__url.host()} - {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "SSL Read Error"
            )
        except ssl.SSLWantWriteError as err:
            logger.debug(f"ssl.SSLWantWriteError for {self.__url.host()} - {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "SSL Write Error"
            )
        except ssl.SSLSyscallError as err:
            logger.debug(f"ssl.SSLSyscallError for {self.__url.host()} - {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "SSL Syscall Error"
            )
        except ssl.SSLEOFError as err:
            logger.debug(f"ssl.SSLEOFError for {self.__url.host()} - {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "SSL EOF Error"
            )
        except ssl.SSLCertVerificationError as err:
            logger.debug(
                f"ssl.SSLCertVerificationError for {self.__url.host()} - {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url,
                RESPONSE_STATUSDETAIL_ERROR_TLS,
                "SSL Certificate Verification Error",
            )
        except ssl.SSLError as err:
            logger.debug(f"ssl.SSLError for {self.__url.host()} - {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "SSL Error"
            )
        except SocketTimeoutException as err:
            logger.debug(
                f"socket.timeout: socket timed out connecting to {self.__url.host()}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_HOST, "Socket timeout"
            )
        except Exception as err:
            logger.error(
                f"Unknown exception encountered when completing SSL handshake for {self.__url.host()} - {err}"
            )
            raise err

    def __validate_ssl_certificate(self, secure_socket) -> CertWrapper:
        """
        Trust-on-first-use (TOFU) validation on SSL certificate or throws exception
        """

        try:
            certificate_wrapper = CertWrapper.parse(secure_socket.getpeercert(True))
            self.__cert_store.validate_tofu_or_add(
                secure_socket.server_hostname, certificate_wrapper
            )
            return certificate_wrapper
        except ValueError as err:
            logger.debug(f"ValueError: {self.__url.netloc()}. {err}")
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, err
            )
        except RemoteCertificateExpired as err:
            logger.debug(
                f"RemoteCertificateExpired: {self.__url.netloc()} has an expired certificate. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_TLS, "Certificate expired"
            )
        except TofuCertificateRejection as err:
            logger.debug(
                f"TofuCertificateRejection: {self.__url.netloc()} has an untrusted, unknown certificate. {err}"
            )
            if self.__raise_errors:
                raise err
            return ResponseFactory.create(
                self.__url,
                RESPONSE_STATUSDETAIL_ERROR_TLS,
                "Untrusted certificate (TOFU rejection)",
            )
        except Exception as err:
            logger.error(
                f"Unknown exception encountered when validating ssl certificate on {self.__url.netloc()} - {err}"
            )
            raise err

    def is_using_ca_cert(self):
        """
        Returns if the request is using ca_cert
        """
        return self.__ca_cert is not None

    def __setup_ssl_default_context(self):
        """
        Setup an SSL default context (without a client certificate)
        This will bypass certificate validation against a CA.
        TOFU validation will be completed after the request is completed.
        """

        context = ssl.create_default_context()
        context.minimum_version = ssl.TLSVersion.TLSv1_2
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        return context

    def __setup_ssl_client_certificate_context(self, context):
        """
        Load cert chain for client certificate
        TODO: Better error handling here?
        """
        cert, key = self.__ca_cert
        context.load_cert_chain(cert, key)

    def __transport_payload(self, socket_obj, payload):
        """
        Handles Gemini protocol negotiation over the socket
        """

        try:
            socket_obj.sendall((f"{payload}{CRLF}").encode(GEMINI_DEFAULT_ENCODING))
            fd = socket_obj.makefile("rb")
            return fd.readline().decode(GEMINI_DEFAULT_ENCODING).strip(), fd.read()
        except SocketTimeoutException:
            logger.debug(
                f"socket.timeout: socket timed out connecting to {self.__url.host()}"
            )
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_HOST, "Socket timeout"
            )
        except Exception as err:
            logger.error(
                f"Unknown exception encountered when transporting data to {self.__url.netloc()} - {err}"
            )
            raise err

    def __handle_response(
        self, header, raw_body, certificate: cryptography.x509.Certificate
    ):
        """
        Handles basic response data from the remote server and hands off to the Response object
        """
        try:
            status, meta = re.split(
                GEMINI_RESPONSE_HEADER_SEPARATOR, header, maxsplit=1
            )

            if not re.match(r"^\d{2}$", status):
                raise GeminiResponseParseError(
                    "Response status is not a two-digit code"
                )

            if len(meta) > GEMINI_RESPONSE_HEADER_META_MAXLENGTH:
                raise GeminiResponseParseError("Header meta text is too long")

            return ResponseFactory.create(
                self.__url, status, meta.strip(), raw_body, certificate
            )
        except GeminiResponseParseError as err:
            return ResponseFactory.create(
                self.__url, RESPONSE_STATUSDETAIL_ERROR_PROTOCOL, err
            )



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import cgi
import logging

from cryptography.x509 import Certificate

from .globals import (
    CRLF,
    GEMINI_DEFAULT_ENCODING,
    GEMINI_DEFAULT_MIME_TYPE,
    RESPONSE_STATUSDETAIL_ERROR_PROTOCOL,
)

logger = logging.getLogger(__name__)


class ResponseFactory:
    """
    Wrapper class for factory:
    Determines the approriate response type based on response status
    and generates the appropriate response type
    """

    @classmethod
    def create(cls, url: str, status: str, meta=None, raw_body=None, certificate=None):
        """
        Given a url, status, and response data, generates the appropriate response type
        """
        basic_status_code = status[0]
        factories = {
            "0": ErrorResponse,
            "1": InputResponse,
            "2": SuccessResponse,
            "3": RedirectResponse,
            "4": TempFailureResponse,
            "5": PermFailureResponse,
            "6": ClientCertRequiredResponse,
        }

        factory_class = factories.get(basic_status_code, None)

        if factory_class is None:
            return ErrorResponse(
                url,
                RESPONSE_STATUSDETAIL_ERROR_PROTOCOL,
                f"Invalid response received from the server, status code: {status}",
                None,
                None,
            )

        return factory_class(url, status, meta, raw_body, certificate)


class BaseResponse:
    """
    Abstract Base response type that all response types inherit from.
    Included public members:
    * url
    * basic_status
    * status
    * meta
    * raw_body
    * certificate
    """

    url: str
    basic_status: str
    status: str
    meta: str
    raw_body: bytes
    certificate: Certificate

    def __init__(
        self,
        url: str,
        status: str,
        meta: str,
        raw_body: bytes,
        certificate: Certificate,
    ):
        """
        Initializes a BaseResponse with the request url, status code, metadata, raw body string, and remote certificate
        """
        self.url = str(url)
        self.basic_status = status[0]
        self.status = status
        self.meta = meta
        self.raw_body = raw_body
        self.certificate = certificate

    def is_a(self, response_class_type):
        """
        Returns true if the response class type matches the current class
        """
        return isinstance(self, response_class_type)

    def success(self):
        """
        Returns true if the response is of the type success
        """
        return self.is_a(SuccessResponse)

    def data(self):
        """
        Fetches processed data from the response.  This method should be overloaded in each specific response type.
        """
        return self.raw_body

    def __str__(self):
        """
        Returns a literal string representation of the response, including messageheader
        """
        return f"{self.status} {self.meta}"

    def __repr__(self):
        """
        A representation of the string for developers
        """
        return f"<ignition.response.{self.__class__.__name__}: {self.status} {self.meta} [from {self.url}]>"


class ErrorResponse(BaseResponse):
    """
    ErrorResponse
    This is a custom response type for ignition, to handle any responses representing request errors
    that are outside of the scope of the Gemini protocol.  Included options are:

    00: RESPONSE_STATUSDETAIL_ERROR_NETWORK
    Any errors that occur at the network level, and prevented the client from making any connection
    with external services.

    01: RESPONSE_STATUSDETAIL_ERROR_DNS = "01"
    Any errors at the DNS level.

    02: RESPONSE_STATUSDETAIL_ERROR_HOST
    Any errors connecting to the host (timeout, refused, etc.).

    03: RESPONSE_STATUSDETAIL_ERROR_TLS
    Any errors associated with TLS/SSL, including handshake errors, certificate expired errors,
    and security errors like certificate rejection errors.

    04: RESPONSE_STATUSDETAIL_ERROR_PROTOCOL
    Any errors where a secure message is received from the server, but it does not conform to the
    Gemini protocol requirements and cannot be processed.
    """

    def data(self):
        """
        Fetch data relevant to the ErrorResponse; in this case the metadata message from the response
        """
        return self.meta


class InputResponse(BaseResponse):
    """
    InputRequest
    Meets Gemini specification: 3.2.1 1x (INPUT)

    Status codes beginning with 1 are INPUT status codes, meaning that
    the requested resource accepts a line of textual user input.

    The user should reissue a request to the url with parameters in the form:
    gemini://hostname/path?query
    """

    def data(self):
        """
        Returns the related instructions for the InputResponse.
        The <META> line is a prompt which should be displayed to the user.
        """
        return self.meta


class SuccessResponse(BaseResponse):
    """
    SuccessResponse
    Meets Gemini specification: 3.2.2 2x (SUCCESS)

    Status codes beginning with 2 are SUCCESS status codes.
    """

    def data(self):
        """
        Decode the success message body using metadata in the appropriate encoding type
        """

        meta = self.meta or GEMINI_DEFAULT_MIME_TYPE
        _, options = cgi.parse_header(meta)
        encoding = (
            options["charset"] if "charset" in options else GEMINI_DEFAULT_ENCODING
        )
        try:
            return self.raw_body.decode(encoding)
        except LookupError:
            logger.warning(
                f"Could not decode response body using invalid encoding {encoding}"
            )
            return self.raw_body
        except UnicodeDecodeError:
            logger.warning(
                f"Could not decode response body via encoding {encoding}, returning raw data"
            )
            return self.raw_body

    def __str__(self):
        """
        The string representation of the success message should be header + body
        """
        return f"{self.status} {self.meta}{CRLF}{self.data()}"


class RedirectResponse(BaseResponse):
    """
    RedirectResponse
    Meets Gemini specification: 3.2.3 3x (REDIRECT)

    Status codes beginning with 3 are REDIRECT status codes.

    The server is redirecting the client to a new location for the requested resource
    """

    def data(self):
        """
        Returns the new destination for redirection from the server
        """
        return self.meta


class TempFailureResponse(BaseResponse):
    """
    TempFailureResponse
    Meets Gemini specification: 3.2.4 4x (TEMPORARY FAILURE)

    Status codes beginning with 4 are TEMPORARY FAILURE status codes.

    The request has failed, but an identical request may success in the future.
    """

    def data(self):
        """
        Returns the data from the server in the META field, which may provide additional information to the user.
        """
        return f"{self.status} {self.meta}"


class PermFailureResponse(BaseResponse):
    """
    PermFailureResponse
    Meets Gemini specification: 3.2.5 5x (PERMANENT FAILURE)

    Status codes beginning with 5 are PERMANENT FAILURE status codes.

    The request has failed, identical requests will likely fail in the future.
    """

    def data(self):
        """
        Returns the data from the server in the META field, which may provide additional information to the user.
        """
        return f"{self.status} {self.meta}"


class ClientCertRequiredResponse(BaseResponse):
    """
    ClientCertRequiredResponse
    Meets Gemini specification: 3.2.6 6x (CLIENT CERTIFICATE REQUIRED)

    Status codes beginning with 6 are CLIENT CERTIFICATE REQUIRED status codes

    The request should be retried with a client certificate.
    """

    def data(self):
        """
        Return additional information from the server on certificate requirements
        or the reason a certificate was rejected
        """
        return self.meta



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

CRLF = "\r\n"
EOL = "\n"

# Gemini-Protocol Mechanical Constants
GEMINI_SCHEME = "gemini"
GEMINI_PORT = 1965
GEMINI_DEFAULT_MIME_TYPE = "text/gemini; charset=utf-8"
GEMINI_DEFAULT_ENCODING = "utf-8"
GEMINI_RESPONSE_HEADER_SEPARATOR = "\\s+"
GEMINI_URL_MAXLENGTH = 1024
GEMINI_RESPONSE_HEADER_META_MAXLENGTH = 1024

# One-character response codes
RESPONSE_STATUS_ERROR = "0"
RESPONSE_STATUS_INPUT = "1"
RESPONSE_STATUS_SUCCESS = "2"
RESPONSE_STATUS_REDIRECT = "3"
RESPONSE_STATUS_TEMP_FAILURE = "4"
RESPONSE_STATUS_PERM_FAILURE = "5"
RESPONSE_STATUS_CLIENTCERT_REQUIRED = "6"

# Two-character response codes
RESPONSE_STATUSDETAIL_ERROR_NETWORK = "00"
RESPONSE_STATUSDETAIL_ERROR_DNS = "01"
RESPONSE_STATUSDETAIL_ERROR_HOST = "02"
RESPONSE_STATUSDETAIL_ERROR_TLS = "03"
RESPONSE_STATUSDETAIL_ERROR_PROTOCOL = "04"
RESPONSE_STATUSDETAIL_INPUT = "10"
RESPONSE_STATUSDETAIL_INPUT_SENSITIVE = "11"
RESPONSE_STATUSDETAIL_SUCCESS = "20"
RESPONSE_STATUSDETAIL_REDIRECT_TEMPORARY = "30"
RESPONSE_STATUSDETAIL_REDIRECT_PERMANENT = "31"
RESPONSE_STATUSDETAIL_TEMP_FAILURE = "40"
RESPONSE_STATUSDETAIL_TEMP_FAILURE_UNAVAILABLE = "41"
RESPONSE_STATUSDETAIL_TEMP_FAILURE_CGI = "42"
RESPONSE_STATUSDETAIL_TEMP_FAILURE_PROXY = "43"
RESPONSE_STATUSDETAIL_TEMP_FAILURE_SLOW_DOWN = "44"
RESPONSE_STATUSDETAIL_PERM_FAILURE = "50"
RESPONSE_STATUSDETAIL_PERM_FAILURE_NOT_FOUND = "51"
RESPONSE_STATUSDETAIL_PERM_FAILURE_GONE = "52"
RESPONSE_STATUSDETAIL_PERM_FAILURE_PROXY_REFUSED = "53"
RESPONSE_STATUSDETAIL_PERM_FAILURE_BAD_REQUEST = "59"
RESPONSE_STATUSDETAIL_CLIENTCERT_REQUIRED = "60"
RESPONSE_STATUSDETAIL_CLIENTCERT_REQUIRED_NOT_AUTHORIZED = "61"
RESPONSE_STATUSDETAIL_CLIENTCERT_REQUIRED_NOT_VALID = "62"

# ignition application defaults
DEFAULT_REQUEST_TIMEOUT = 30
DEFAULT_HOSTS_FILE = ".known_hosts"



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import datetime

from ..exceptions import CertRecordParseException
from ..globals import EOL


class CertRecord:
    """
    Manages a single Certificate Record with a hostfile, signature, and expiration
    """

    hostname: str
    fingerprint: str
    expiration: datetime.datetime

    def __init__(self, hostname: str, fingerprint: str, expiration: datetime.datetime):
        """
        Generate a CertRecord from logic; passing in the hostname, fingerprint, and expiration
        """
        self.hostname = hostname
        self.fingerprint = fingerprint
        self.expiration = expiration

    @classmethod
    def from_string(cls, host_string: str):
        """
        Generate a CertRecord from a string in the format:
        [HOSTNAME] [SSH-ALGORITHM PUBLIC_KEY];EXPIRES=[YYYY-MM-DDTHH:mm:ss.SSSZ]
        """
        try:
            hostname, fingerprint_with_expiration = host_string.strip().split(
                " ", maxsplit=1
            )
            fingerprint, expiration = fingerprint_with_expiration.split(";EXPIRES=")
            expiration_datetime = datetime.datetime.fromisoformat(expiration)

            return CertRecord(hostname, fingerprint, expiration_datetime)
        except Exception as e:
            raise CertRecordParseException() from e

    def to_string(self):
        """
        Converts a CertRecord to a string in the format:
        [HOSTNAME] [SSH-ALGORITHM PUBLIC_KEY];EXPIRES=[YYYY-MM-DDTHH:mm:ss.SSSZ]
        """
        return (
            self.hostname
            + " "
            + self.fingerprint
            + ";EXPIRES="
            + self.expiration.isoformat()
            + EOL
        )

    def is_expired(self):
        """
        Returns true if the expiration date on the cert record is before now
        """
        return self.expiration < self.now()

    def now(self):
        """
        Utility function to get current datetime, extracted for testing purposes
        Returns datetime
        """
        return datetime.datetime.now()



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

from .globals import *
from .request import Request
from .response import (
    ClientCertRequiredResponse,
    ErrorResponse,
    InputResponse,
    PermFailureResponse,
    RedirectResponse,
    SuccessResponse,
    TempFailureResponse,
)
from .ssl.cert_store import CertStore
from .util import TimeoutManager

__version__ = "0.2.1"

__timeout = TimeoutManager(DEFAULT_REQUEST_TIMEOUT)
__cert_store = CertStore(DEFAULT_HOSTS_FILE)


def set_default_hosts_file(hosts_file):
    """
    Set the default host file location where all of the certificate fingerprints
    are stored in order to support Trust-On-First-Use (TOFU) validation.
    By default, this file is stored in the root directory as your project in a
    file named `.known_hosts`. This can be updated to any readable location but
    should be stored somewhere persistent for security purposes.

    The format of this file is very similar to (but not identical to)
    the SSH `known_hosts` file.

    Parameters:
    * hosts_file: `string`

    """
    __cert_store.set_hosts_file(hosts_file)


def set_default_timeout(timeout):
    """
    Set the default timeout (in seconds) for all requests made via ignition.
    The default timeout is 30 seconds.

    Parameters:
    * timeout: `float`
    """
    __timeout.set_default_timeout(timeout)


def url(request_url, referer=None):
    """
    Given a *url* to a Gemini capsule, this returns a standardized,
    fully-qualified url to the Gemini capsule.  If a *referer* is
    provided, a dynamic URL is constructed by ignition to send a request
    to.  This logic follows URL definition behavior outlined in
    [RFC-3986](https://tools.ietf.org/html/rfc3986).

    This allows for the bulk of URL generation logic to be handled
    without ignition as opposed to within the business logic of the client.
    Here are some sample use cases:

    *Use Case 1: Automatically populate URL protocol*
    ```python
    ignition.url('//geminiprotocol.net') # => gemini://geminiprotocol.net
    ```

    *Use Case 2: Navigate to an absolute path*
    ```python
    ignition.url('/home', 'gemini://geminiprotocol.net') # => gemini://geminiprotocol.net/home
    ```

    *Use Case 3: Navigate to a relative path*
    ```python
    ignition.url('2', 'gemini://geminiprotocol.net/home') # => gemini://geminiprotocol.net/home/2
    ```

    *Use Case 4: Resolve paths with navigation*
    ```python
    ignition.url('../fun/', 'gemini://geminiprotocol.net/home/work/') # => gemini://geminiprotocol.net/home/fun/
    ```

    *Note:* if the user's intent is to generate a url to a Gemini capsule and then make a request,
    ignition recommends that you just provide the *url* and *referer* to `ignition.request()`, as
    that function encapsulates all of the logic within this method when making a request.  If you
    want to retrieve a URL from an already processed request, it is recommended to use
    `ignition.BaseResponse.url`, as that will store the URL that was actually used.  This method
    is only intended for use in constructing a URL but not generating a request.

    Parameters:
    * url: `string`
    * referer: `string` (optional)

    Returns: `string`
    """
    dummy_req = Request(request_url, referer=referer)
    return dummy_req.get_url()


def request(request_url, referer=None, timeout=None, raise_errors=False, ca_cert=None):
    """
    Given a *url* to a Gemini capsule, this performs a request to the specified
    url and returns a response (as a subclass of [ignition.BaseResponse](#ignitionbaseresponse))
    with the details associated to the response.  This is the interface that most
    users should use.

    If a *referer* is provided, a dynamic URL is constructed by ignition to send a
    request to. (*referer* expectes a fully qualified url as returned by
    `ignition.BaseResponse.url` or (less prefered) `ignition.url()`).
    Typically, in order to simplify the browsing experience, you should pass
    the previously requested URL as the referer to simplify URL construction logic.

    *See `ignition.url()` for details around url construction with a referer.*

    If *raise_errors* is `True` (default value = `False`), then non-protocol errors
    will bubble up and be raised as an exception instead of returning
    [ignition.ErrorResponse](#ignitionerrorresponse)

    If a *timeout* is provided, this will specify the client timeout (in seconds)
    for this request.  The default is 30 seconds.  See also `ignition.set_default_timeout`
    to change the default timeout.

    If a *ca_cert* is provided, the certificate will be sent to the server as a CA CERT.
    You will need to provide the paths to both the certificate and the key in this case.

    Depending on the response from the server, as per Gemini specification, the
    corresponding response type will be returned.

    * If the response status begins with "1", the response type is `INPUT`,
      and will return a response of type [ignition.InputResponse](#ignitioninputresponse).
    * If the response status begins with "2", the response type is `STATUS`,
      and will return a response of type [ignition.SuccessResponse](#ignitionsuccessresponse).
    * If the response status begins with "3", the response type is `REDIRECT`,
      and will return a response of type [ignition.RedirectResponse](#ignitionredirectresponse).
    * If the response status begins with "4", the response type is `TEMPORARY FAILURE`,
      and will return a response of type [ignition.TempFailureResponse](#ignitiontempfailureresponse).
    * If the response status begins with "5", the response type is `PERMANENT FAILURE`,
      and will return a response of type [ignition.PermFailureResponse](#ignitionpermfailureresponse).
    * If the response status begins with "6", the response type is `CLIENT CERTIFICATE REQUIRED`,
      and will return a response of type [ignition.ClientCertRequiredResponse](#ignitionclientcertrequiredresponse).
    * If *no valid response* can be returned, ignition assigns a response type of "0"
      and returns a response of type [ignition.ErrorResponse](#ignitionerrorresponse).
      Note: if the user specifies `raise_errors=True`, then any errors will bubble up instead of returning
      this response type.

    Parameters:
    * url: `string`
    * referer: `string` (optional)
    * timeout: `float` (optional)
    * raise_errors: `bool` (optional)
    * ca_cert: `Tuple(cert_file, key_file)` (optional)

    Returns: `[ignition.BaseResponse](#ignitionbaseresponse)`
    """

    req = Request(
        request_url,
        cert_store=__cert_store,
        request_timeout=__timeout.get_timeout(timeout),
        referer=referer,
        ca_cert=ca_cert,
        raise_errors=raise_errors,
    )

    return req.send()


__all__ = [
    "set_default_hosts_file",
    "set_default_timeout",
    "url",
    "request",
    "ClientCertRequiredResponse",
    "ErrorResponse",
    "InputResponse",
    "PermFailureResponse",
    "RedirectResponse",
    "SuccessResponse",
    "TempFailureResponse",
    "RESPONSE_STATUS_ERROR",
    "RESPONSE_STATUS_INPUT",
    "RESPONSE_STATUS_SUCCESS",
    "RESPONSE_STATUS_REDIRECT",
    "RESPONSE_STATUS_TEMP_FAILURE",
    "RESPONSE_STATUS_PERM_FAILURE",
    "RESPONSE_STATUS_CLIENTCERT_REQUIRED",
    "RESPONSE_STATUSDETAIL_ERROR_NETWORK",
    "RESPONSE_STATUSDETAIL_ERROR_DNS",
    "RESPONSE_STATUSDETAIL_ERROR_HOST",
    "RESPONSE_STATUSDETAIL_ERROR_TLS",
    "RESPONSE_STATUSDETAIL_ERROR_PROTOCOL",
    "RESPONSE_STATUSDETAIL_INPUT",
    "RESPONSE_STATUSDETAIL_INPUT_SENSITIVE",
    "RESPONSE_STATUSDETAIL_SUCCESS",
    "RESPONSE_STATUSDETAIL_REDIRECT_TEMPORARY",
    "RESPONSE_STATUSDETAIL_REDIRECT_PERMANENT",
    "RESPONSE_STATUSDETAIL_TEMP_FAILURE",
    "RESPONSE_STATUSDETAIL_TEMP_FAILURE_UNAVAILABLE",
    "RESPONSE_STATUSDETAIL_TEMP_FAILURE_CGI",
    "RESPONSE_STATUSDETAIL_TEMP_FAILURE_PROXY",
    "RESPONSE_STATUSDETAIL_TEMP_FAILURE_SLOW_DOWN",
    "RESPONSE_STATUSDETAIL_PERM_FAILURE",
    "RESPONSE_STATUSDETAIL_PERM_FAILURE_NOT_FOUND",
    "RESPONSE_STATUSDETAIL_PERM_FAILURE_GONE",
    "RESPONSE_STATUSDETAIL_PERM_FAILURE_PROXY_REFUSED",
    "RESPONSE_STATUSDETAIL_PERM_FAILURE_BAD_REQUEST",
    "RESPONSE_STATUSDETAIL_CLIENTCERT_REQUIRED",
    "RESPONSE_STATUSDETAIL_CLIENTCERT_REQUIRED_NOT_AUTHORIZED",
    "RESPONSE_STATUSDETAIL_CLIENTCERT_REQUIRED_NOT_VALID",
]



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import sys

# Polyfill to include gemini in urllib parsing
if sys.version_info > (3, 13):
    raise Exception("Python versions > 3.12.x are not supported at this time.")
if sys.version_info > (3, 12):
    from .python3_12.Lib import urllib
if sys.version_info > (3, 11):
    from .python3_11.Lib import urllib
if sys.version_info > (3, 10):
    from .python3_10.Lib import urllib
if sys.version_info > (3, 9):
    from .python3_9.Lib import urllib
elif sys.version_info > (3, 8):
    from .python3_8.Lib import urllib
elif sys.version_info > (3, 7):
    from .python3_7.Lib import urllib
else:
    raise Exception("Python versions < 3.7 are not supported at this time.")



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

from . import __version__

if __name__ == "__main__":
    print(__version__)



"""
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
"""

import logging

from .globals import GEMINI_PORT, GEMINI_SCHEME
from .python import urllib
from .util import normalize_path

logger = logging.getLogger(__name__)


class URL:
    """
    The URL class negotiates the correct URL based on passed in URL.

    This logic prepares the URL to be passed via the socket connector,
    as well as for the data payload for Gemini.
    """

    def __init__(self, url, referer_url=None):
        """
        Construct a protocool-safe URL based on the passed string.
        """
        self.__parsed_url = self.__url_constructor(url, referer_url)

        logger.debug(
            (
                f"Recieved url {url} for parsing, {f'with referer {referer_url}, ' if referer_url else ''} generated gemini url: {self}"
            )
        )

    def __url_constructor(self, url, referer_url):
        """
        Constructs a protocol-safe URL based on the passed string.

        If referer_url is included (which should be the constructed
        URL from the last time this ran), the new url is joined onto
        the referer.  This allows the user to pass in paths without a
        hostname.
        """

        url = url.lstrip()
        base_url = url
        if referer_url:
            base_url = urllib.parse.urljoin(referer_url, url, False)

        return urllib.parse.urlsplit(base_url, GEMINI_SCHEME, False)

    def __str__(self):
        """
        Custom logic to re-join the URL into a string
        TODO url = 'about:blank', 'example:test' RFC-6694 and RFC-7585
        """

        return "".join(
            [
                self.protocol(),
                self.host(),
                (f":{self.port()}" if self.port() != GEMINI_PORT else ""),
                self.path(),
                (f"?{self.query()}" if self.query() else ""),
            ]
        )

    def path(self):
        """
        Returns path portion of the url
        URL Schema: scheme://host:port/path?query
        """

        return normalize_path(self.__parsed_url.path or "")

    def host(self):
        """
        Returns host portion of the url
        URL Schema: scheme://host:port/path?query
        """

        return self.__parsed_url.hostname or ""

    def port(self):
        """
        Returns port portion of the url
        URL Schema: scheme://host:port/path?query
        """

        try:
            return self.__parsed_url.port or GEMINI_PORT
        except ValueError:
            # https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlsplit
            logger.warning(
                f"There was an error reading the port from the url. Defaulting to {GEMINI_PORT}"
            )
            return GEMINI_PORT

    def netloc(self):
        """
        Returns netloc portion of the url, which is the host:port
        URL Schema: scheme://host:port/path?query
        """

        return self.__parsed_url.netloc

    def protocol(self):
        """
        Returns scheme portion of the url with the protocol designator "://"
        URL Schema: scheme://host:port/path?query
        """

        return f"{self.__parsed_url.scheme}://"

    def query(self):
        """
        Returns query portion of the url
        URL Schema: scheme://host:port/path?query
        """
        return self.__parsed_url.query
